use super::*;
use std::vec::Vec;
use renderer::*;
use std::cmp::{min, max};
use super::animation::*;

use rand::seq::SliceRandom;

pub const ANIMATION_TICK_TIME : u64 = 50;
pub const CAMERA_SLUGGISHNESS : f32 = 0.995;

pub struct TileProperties {
    passable : bool,
    translucent : bool,
    default_tile : (u8, u8),
}

const TILE_PLAINS : u8 = 2;
const TILE_FLAME : u8 = 3;
const TILE_DIM_PUMPKIN : u8 = 5;
const TILE_LIT_PUMPKIN : u8 = 6;
const TILE_IMPASSIBLE_GRASS : u8 = 7;
pub const TILE_TEMPLATES : [TileProperties; 8] = [
    TileProperties {
        passable : false,
        translucent : false,
        default_tile : (0, 0),
    },
    TileProperties {
        passable : true,
        translucent : true,
        default_tile : (1, 0),
    },
    TileProperties {
        passable : true,
        translucent : true,
        default_tile : (2, 0),
    },
    TileProperties {
        passable : true,
        translucent : true,
        default_tile : (3, 0),
    },
    TileProperties {
        passable : false,
        translucent : true,
        default_tile : (12, 0),
    },
    TileProperties {
        passable : false,
        translucent : true,
        default_tile : (13, 0),
    },
    TileProperties {
        passable : false,
        translucent : true,
        default_tile : (14, 0),
    },
    TileProperties {
        passable : false,
        translucent : true,
        default_tile : (1, 0),
    },
];

fn tile_index(tile: &TileProperties) -> u8 {
    let templates = TILE_TEMPLATES.as_ptr() as usize;
    let element = tile as *const TileProperties as usize;
    ((element - templates) / std::mem::size_of::<TileProperties>()) as u8
}

#[derive(Clone, Copy)]
struct FloofTile {
    tile : &'static TileProperties,
    light : u8,
    light_source : u8,
}

impl FloofTile {
    fn new(tile_number : usize) -> Self {
        FloofTile {
            tile : &TILE_TEMPLATES[tile_number],
            light : 0,
            light_source :0,
        }
    }
}

struct Player {
    position : (u16, u16),
    moving : bool,
    struggling : bool,
    controls : Option<Direction>,
    animation : Animation,
    flames : u8,
    lifes : u8,
}

pub struct Game {
    board : std::vec::Vec<FloofTile>,
    width : u16,
    height : u16,
    time : u64,
    last_animation_tick : u64,
    player : Player,
    flames: Vec<Flame>,
    monsters: Vec<Monster>,
    pumpkins_remaining: u16,
    ui_pumpkin: Animation,
    ui_flame: Animation,
}

pub struct Monster {
    position: (u16, u16),
    moving: bool,
    dying : bool,
    wander_direction : Option<Direction>,
    animation : Animation,
}

pub struct Flame {
    animation : Option<Animation>,
    position: (u16, u16),
}

const LIGHT_DECREASE : u8 = 32;
const DIAGONAL_LIGHT_DECREASE : u8 = 44;

impl Game {
    pub fn new(floor : &[u8], width : u16, height : u16) -> Option<Self> {
        let size = width as usize * height as usize;
        if size != floor.len() {return None};
        let mut board = Vec::<FloofTile>::with_capacity(size);
        let mut flames = Vec::new();
        let mut pumpkins_remaining = 0;
        for y in 0..(height as usize) {
            for x in 0..(width as usize) {
                let y_image = height as usize - y - 1;
                let a = floor[x + y_image*width as usize];
                board.push(FloofTile::new(a as usize));
                if a == TILE_FLAME  {
                    let flame = Flame {
                        position : (x as u16, y as u16),
                        animation : None,
                    };
                    flames.push(flame);
                } else if a == TILE_DIM_PUMPKIN {
                    pumpkins_remaining += 1;
                }
            }
        }
        let monsters = Vec::new();
        let time = 0;
        let last_animation_tick = time;
        let player = Player {
            position : (70, height-41),
            moving : false,
            struggling : false,
            controls : None,
            animation : Animation::new(ANIM_PLAYER_STANDS_STILL, true),
            flames : 0,
            lifes : 1,
            };
        let mut game = Game {
            board,
            width : width,
            height : height,
            time,
            last_animation_tick,
            player,
            flames,
            monsters,
            pumpkins_remaining,
            ui_pumpkin: Animation::new(ANIM_UI_PUMPKIN, true),
            ui_flame: Animation::new(ANIM_FLAME, true),
        };
        for i in 0..game.flames.len() {
            game.light_flame(i);
        }
        Some(game)
    }
    fn light_flame(&mut self, i: usize) {
        if let Some(flame) = self.flames.get_mut(i) {
            flame.animation = Some(Animation::new(ANIM_FLAME, true));
            let coords = flame.position;
            self.set_light_source(coords, 128);
        }
    }
    pub fn tile_index(&self, x : u16, y : u16) -> Option<usize> {
        if x >= self.width {
            None
        } else if y >= self.height {
            None
        } else {
            Some(x as usize + y as usize * self.width as usize)
        }
    }
    pub fn floor_tile (&self, coords: (u16, u16)) -> Option<&FloofTile> {
        let (x, y) = coords;
        self.tile_index(x,y).map(|i| &self.board[i])
    }
    pub fn mut_floor_tile (&mut self, coords: (u16, u16)) -> Option<&mut FloofTile> {
        let (x, y) = coords;
        self.tile_index(x,y).map(move |i| {&mut self.board[i]})
    }
    fn propagate_light_increase(&mut self, coords: (u16, u16), new_light: u8) {
        let (x, y) = coords;
        {
            let tile = if let Some(x) = self.mut_floor_tile(coords) {x} else {return};
            if tile.light >= new_light {return};
            if !tile.tile.translucent {return};
            tile.light = new_light;
        }
        if new_light > DIAGONAL_LIGHT_DECREASE {
            let diagonal_light = new_light - DIAGONAL_LIGHT_DECREASE;
            self.propagate_light_increase((x - 1, y - 1), diagonal_light);
            self.propagate_light_increase((x + 1, y - 1), diagonal_light);
            self.propagate_light_increase((x - 1, y + 1), diagonal_light);
            self.propagate_light_increase((x + 1, y + 1), diagonal_light);
        }
        if new_light > LIGHT_DECREASE {
            let straight_light = new_light - LIGHT_DECREASE;
            self.propagate_light_increase((x, y - 1), straight_light);
            self.propagate_light_increase((x - 1, y), straight_light);
            self.propagate_light_increase((x + 1, y), straight_light);
            self.propagate_light_increase((x, y + 1), straight_light);
        }
    }
    fn propagate_light_decrease(&mut self, coords: (u16, u16)) {
        let (x, y) = coords;
        {
            let tile = if let Some(x) = self.floor_tile(coords) {x} else {return};
            let mut new_light = tile.light_source;
            let propagate_further = tile.tile.translucent;
            let mut check_light = |x, y, decrease| {
                if let Some(tile) = self.floor_tile((x, y)) {
                    if tile.tile.translucent && tile.light > decrease {
                        new_light = max(new_light, tile.light - decrease);
                    }
                }
            };
            check_light(x, y - 1, LIGHT_DECREASE);
            check_light(x - 1, y, LIGHT_DECREASE);
            check_light(x + 1, y, LIGHT_DECREASE);
            check_light(x, y + 1, LIGHT_DECREASE);
            check_light(x - 1, y - 1, DIAGONAL_LIGHT_DECREASE);
            check_light(x + 1, y - 1, DIAGONAL_LIGHT_DECREASE);
            check_light(x - 1, y + 1, DIAGONAL_LIGHT_DECREASE);
            check_light(x + 1, y + 1, DIAGONAL_LIGHT_DECREASE);
            if tile.light > new_light {
                self.mut_floor_tile(coords).unwrap().light = new_light;
                if propagate_further {
                    self.propagate_light_decrease((x, y - 1));
                    self.propagate_light_decrease((x - 1, y));
                    self.propagate_light_decrease((x + 1, y));
                    self.propagate_light_decrease((x, y + 1));
                    self.propagate_light_decrease((x - 1, y - 1));
                    self.propagate_light_decrease((x + 1, y - 1));
                    self.propagate_light_decrease((x - 1, y + 1));
                    self.propagate_light_decrease((x + 1, y + 1));
                }
            }
        }
    }
    fn set_light_source(&mut self, coords: (u16, u16), light: u8) {
        let (x, y) = coords;
        let tile = if let Some(x) = self.mut_floor_tile(coords) {x} else {return};
        if !tile.tile.translucent {return};
        let old_light = tile.light;
        tile.light_source = light;
        if old_light < light {
            self.propagate_light_increase(coords, light);
        } else if old_light > light {
            self.propagate_light_decrease(coords);
        }
    }
    fn can_move_in_direction(&self, coords: (u16, u16), dir: Direction) -> Option<(u16, u16)> {
        let new_coords = dir.from(coords);
        if let Some(tile) = self.floor_tile(new_coords) {
            if tile.tile.passable {Some(new_coords)}
            else {None}
        } else {None}
    }

    fn tick(&mut self) {
        self.ui_pumpkin.tick();
        self.ui_flame.tick();
        self.flames.iter_mut().for_each( |a| {
            if let Some(a) = a.animation.as_mut() {
                a.tick();
            }
        });
        self.monsters.iter_mut().for_each(|m| {
            m.animation.tick();
        });
        self.try_spawn_monsters();
        self.player.animation.tick();
        if self.player.animation.done() {
            self.player.moving = false;
            self.player.struggling = false;
        }
        if (!self.player.moving) {
            if(
                self.player_collect_magic_flame() ||
                self.player_attack_monster() ||
                self.start_moving_player() ||
                false){};
        }
        self.monsters_move();
    }
    fn time_passed(&mut self, time: u64) -> u64 {
        let passed = time - self.time;
        self.time = time;
        // something has taken very long. Maybe the window was inactive?
        if passed > 1000 {
            log(format!{"{} ms of time have passed all at once!!!", passed}.as_str());
            self.last_animation_tick = time;
            return 0;
        }
        loop {
            let tick_time = self.last_animation_tick - time;
            if tick_time > ANIMATION_TICK_TIME {
                self.last_animation_tick += ANIMATION_TICK_TIME;
                self.tick();
            } else {
                break;
            }
        }
        passed
    }
    pub fn render (&mut self, buffer : &mut Vec<Vertex>, time: u64) -> (renderer::CameraView, f32) { 
        let dt = self.time_passed(time);
        let (x, y) = self.player.position;
        let render_radius = 32;
        let lowx = if x > render_radius {x - render_radius} else {0};
        let highx = x + render_radius;
        let lowy = if y > render_radius {y - render_radius} else {0};
        let highy = y + render_radius;
        for y in lowy..highy as u16 {
            for x in lowx..highx as u16{
                self.render_tile(buffer, (x, y));
            }
        }
        for a in &mut self.flames {
            let coords = a.position;
            if let Some(a) = a.animation.as_mut() {
                a.render(buffer, coords, (255, 255, 255, 255));
            }
        }
        self.monsters.iter().for_each( |m| {
            let tile = self.floor_tile(m.position).unwrap();
            let l = tile.light;
            m.animation.render(buffer, m.position, (l,l,l,l));
        });

        {
            let tile = self.floor_tile(self.player.position).unwrap();
            let pl = if self.player.flames > 0 {255} else {max(tile.light, 1)};
            self.player.animation.render(buffer, self.player.position, (pl, pl, pl, pl));
        }
    
        let new_view;
        {
            let view_radius = if self.pumpkins_remaining > 0 {max(self.player_light() / 2, 16)}
                    else {128};
            let (x, y) = if self.pumpkins_remaining > 0 {self.player.position}
                    else {(self.width/2, self.height/2)};
            let x = x as f32 * 16.0 + 8.0;
            let y = y as f32 * 16.0 + 8.0;
            new_view = renderer::CameraView::new(view_radius as f32, (x, y));
        }
        (new_view, CAMERA_SLUGGISHNESS.powi(dt as i32))
    }
    pub fn render_ui (&mut self, buffer : &mut Vec<Vertex>) { 
        let mut x = 0;
        for _ in 1..self.player.lifes {
            self.ui_pumpkin.render(buffer, (x, 0), (255, 255, 255, 255));
            x+=1;
        }
        for _ in 1..self.player.flames {
            self.ui_flame.render(buffer, (x, 0), (255, 255, 255, 255));
            x+=1;
        }
    }
    fn render_tile(&self, data : &mut Vec<Vertex>, coords : (u16, u16)) {
        let (x, y) = coords;
        let tile = if let Some(x) = self.floor_tile(coords) {x} else {return};
        let cx = if tile.tile.translucent {1} else {0};
        let light = tile.light;
        let get_light = |x, y| {
            if let Some(tile) = self.floor_tile((x, y)) {
                (tile.light, if tile.tile.translucent {1} else {0})
            } else {
                (0, 0)
            }
        };
        let (bl, blx) = get_light(x - 1, y - 1);
        let (br, brx) = get_light(x + 1, y - 1);
        let (ul, ulx) = get_light(x - 1, y + 1);
        let (ur, urx) = get_light(x + 1, y + 1);
        let (b, bx) = get_light(x, y - 1);
        let (l, lx) = get_light(x - 1, y);
        let (r, rx) = get_light(x + 1, y);
        let (u, ux) = get_light(x, y + 1);
        let mean4 = |a, b, c, d, x| {
            if x > 0 
                {((a as u16 + b as u16 + c as u16 + d as u16) / x) as u8}
            else
                {0}
        };
        let lbl = mean4(light, b, l, bl, cx + bx + lx + blx);
        let lul = mean4(light, u, l, ul, cx + ux + lx + ulx);
        let lur = mean4(light, u, r, ur, cx + ux + rx + urx);
        let lbr = mean4(light, b, r, br, cx + bx + rx + brx);
        define_rect(
                data, 
                (x * 16, y * 16),
                tile.tile.default_tile,
                (lbl, lul, lur, lbr));
    }
    pub fn on_mouse_click(&mut self, renderer : &mut Renderer, pos : (f32, f32)) {
    /*
        let model_pos = renderer.model_position_from_screen_coords(pos);
        let (x, y) = model_pos;
        let x = x/16.0;
        let y = y/16.0;
        let tile_clicked = if
                x >= 0.0 && x < self.width as f32 &&
                y >= 0.0 && y < self.height as f32 {
            Some((x as u16, y as u16))
        } else {None};

        if let Some((x, y)) = tile_clicked {
            self.move_player_to(x, y);
            log(format!("moved player to {} {}",x, y).as_str());
        } else {
            log("Clicked off the board");
        }
    */
    }
    pub fn user_input(&mut self, direction: Option<Direction>) {
        self.player.controls = direction;
    }
    fn monsters_move(&mut self) {
        for i in (0..self.monsters.len()).rev() {
            if self.monsters[i].animation.done() && self.monsters[i].dying {
                self.monsters.swap_remove(i);
            }
        }
        for i in 0..self.monsters.len() {
            if self.monsters[i].animation.done() {
                let coords = self.monsters[i].position;
                let mut direction = self.monsters[i].wander_direction.or_else(|| Some(Direction::random_cardinal())).unwrap();
                if rand::random::<u32>() % 13 == 0 {
                    direction = Direction::random_cardinal();
                }
                if let Some(new_coords) = self.can_move_in_direction(coords, direction) {
                    self.monsters[i].wander_direction = Some(direction);
                    let player_coords = self.player.position;
                    if (new_coords == player_coords) {
                        if (!self.player.struggling) {
                            self.monster_attack_player(i, direction);
                        }
                    } else {
                        self.monsters[i].animation = Animation::new(
                            match direction {
                                Direction::Left => ANIM_MONSTER_WALKS_LEFT,
                                Direction::Right => ANIM_MONSTER_WALKS_RIGHT,
                                Direction::Down => ANIM_MONSTER_WALKS_DOWN,
                                _ => ANIM_MONSTER_WALKS_UP,
                            }, false);
                        self.monsters[i].position = new_coords;
                    }
                } else {
                    self.monsters[i].wander_direction = None;
                }
            }
        }
    }

    fn monster_attack_player(&mut self, i: usize, direction :Direction) {
        self.monsters[i].animation = Animation::new(ANIM_MONSTER_ATTACKS, false);
        self.player.moving = true;
        self.player.struggling = true;
        self.player.animation = Animation::new(ANIM_PLAYER_STRUGGLE, false);
        self.player.lifes -= 1;
        if (self.player.lifes == 0) {
            // game over!
            self.player.flames = 1;
            self.set_light_source(self.player.position, 0);
            self.player.position = (1, 1);
            self.set_light_source((1,2), 255);
            self.set_light_source((1,1), 255);
        }
    }
    
    fn player_attack_monster(&mut self) -> bool {
        let coords = self.player.position;
        if let Some(monster_coords) = self.player.controls.map(|x| x.from(coords)) {
            if let Some(monster) = self.monsters.iter_mut().find(|x| x.position == monster_coords) {
                if !monster.dying {
                    monster.dying = true;
                    monster.animation = Animation::new(ANIM_MONSTER_DESTROYED, false);
                }
                return true;
            }
        }
        false
    }

    fn player_collect_magic_flame(&mut self) -> bool {
        // player steps on a flame tile?
        let coords = self.player.position;
        for i in 0..self.flames.len() {
            if self.flames[i].animation.is_some() && self.flames[i].position == coords {
                self.flames[i].animation = None;
                self.player.flames += 1;
                self.player.animation = Animation::new(ANIM_PLAYER_AFLAME, false);
                self.player.moving = true;
                self.set_light_source(coords, self.player_light());
                return true;
            }
        }
        false
    }

    fn start_moving_player(&mut self) -> bool {
        if let Some(direction) = self.player.controls {
            let coords = direction.from(self.player.position);
            if let Some(tile) = self.floor_tile(coords) {
                if tile.tile.passable {
                    let light_source = tile.light_source;
                    let player_light = self.player_light();
                    self.player.animation = if self.player.flames > 0 {
                        match direction {
                            Direction::Left => {Animation::new(ANIM_LIT_PLAYER_WALKS_LEFT, false)},
                            Direction::Right => {Animation::new(ANIM_LIT_PLAYER_WALKS_RIGHT, false)},
                            Direction::Down => {Animation::new(ANIM_LIT_PLAYER_WALKS_DOWN, false)},
                            Direction::Up => {Animation::new(ANIM_LIT_PLAYER_WALKS_UP, false)},
                        _ => {return false}}
                    } else {
                        match direction {
                            Direction::Left => {Animation::new(ANIM_PLAYER_WALKS_LEFT, false)},
                            Direction::Right => {Animation::new(ANIM_PLAYER_WALKS_RIGHT, false)},
                            Direction::Down => {Animation::new(ANIM_PLAYER_WALKS_DOWN, false)},
                            Direction::Up => {Animation::new(ANIM_PLAYER_WALKS_UP, false)},
                        _ => {return false}}
                    };
                    if light_source < player_light {
                        self.set_light_source(coords, self.player_light());
                    }
                    self.set_light_source(self.player.position, 0);
                    self.player.moving = true;
                    self.player.position = coords;
                    return true;
                } else if tile_index(tile.tile) == TILE_DIM_PUMPKIN {
                    if self.player.flames > 1 {
                        let tile = self.mut_floor_tile(coords).unwrap();
                        tile.tile = &TILE_TEMPLATES[TILE_LIT_PUMPKIN as usize];
                        self.pumpkins_remaining -= 1;
                        self.player.flames -= 1;
                        self.player.lifes += 1;
                        self.set_light_source(self.player.position, self.player_light());
                        self.set_light_source(coords, 255);
                        if self.pumpkins_remaining == 0 {
                            self.monsters.clear();
                            self.winning_shenenigans();
                        }
                    }
                }
            }
        };
        false
    }

    fn winning_shenenigans(&mut self) {
        for y in 0..self.height {
            for x in 0..self.width {
                if let Some(tile) = self.mut_floor_tile((x, y)) {
                    if tile_index(tile.tile) == TILE_LIT_PUMPKIN {
                        tile.tile = &TILE_TEMPLATES[TILE_IMPASSIBLE_GRASS as usize];
                        let mut animation = Animation::new(ANIM_VICTORY, true);
                        let n = rand::random::<u32>() % ANIM_VICTORY.len() as u32;
                        for _ in 0..n {
                            animation.tick();
                        }
                        self.flames.push(Flame{
                            animation: Some(animation),
                            position: (x, y),
                        });
                    }
                }
            }
        }
    }

    fn player_light(&self) -> u8 {
        let flames = self.player.flames;
        if flames == 0 {0} else {
            let light = flames as u16 * 16 + 32;
            if light >= 256 {
                255
            } else {
                light as u8
            }
        }
    }

    fn try_spawn_monsters(&mut self) {
        if self.pumpkins_remaining == 0 {return};
        let flames = self.player.flames;
        let monster_count = flames as usize * 6;
        if self.monsters.len() < monster_count {
            let x = rand::random::<u16>() % self.width;
            let y = rand::random::<u16>() % self.height;
            let tile = self.floor_tile((x, y)).unwrap();
            let tile_index = tile_index(tile.tile);
            if tile_index == TILE_PLAINS {
                let monster = Monster {
                    position : (x, y),
                    moving : false,
                    dying : false,
                    wander_direction : None,
                    animation : Animation::new(ANIM_MONSTER_RISES, false),};
                self.monsters.push(monster);
            } else {
            }
        }
    }
}

#[derive(Clone, Copy)]
pub enum Direction {
    Left,
    Right,
    Down,
    Up,
    LeftDown,
    LeftUp,
    RightDown,
    RightUp,
}

impl Direction {
    fn from(&self, pos: (u16, u16)) -> (u16, u16){
        let (x, y) = pos;
        let add = |a: u16, b: u16| {let (r, _) = a.overflowing_add(b); r};
        let sub = |a: u16, b: u16| {let (r, _) = a.overflowing_sub(b); r};
        match self {
            Direction::Left =>      (sub(x,1),  y),
            Direction::Right =>     (add(x,1),  y),
            Direction::Down =>      (x,         sub(y,1)),
            Direction::Up =>        (x,         add(y,1)),
            Direction::LeftDown =>  (sub(x,1),  sub(y,1)),
            Direction::LeftUp =>    (sub(x,1),  add(y,1)),
            Direction::RightDown => (add(x,1),  sub(y,1)),
            Direction::RightUp =>   (add(x,1),  add(y,1)),
        }
    }
    fn random_cardinal() -> Self {
        let x = rand::random::<i32>() % 4;
        match(x) {
            0 => Direction::Left,
            1 => Direction::Right,
            2 => Direction::Down,
            _ => Direction::Up,
        }
    }
}

