use wasm_bindgen::prelude::*;
use std::cell::RefCell;
use std::rc::Rc;
use wasm_bindgen::{JsCast};
use web_sys::{
    WebGlRenderingContext as Gl,
    WebGlShader as GlShader,
    WebGlProgram as GlProgram,
//    WebGlVertexArrayObject as GlVertexArray,
    WebGlBuffer as GlBuffer,
    WebGlUniformLocation as GlUniformLocation,
    WebGlTexture as GlTexture,
};
use nalgebra_glm as glm;
use field_offset::offset_of;
use png;
use rand;

mod shader;
mod renderer;
mod game;
mod animation;

macro_rules! install_cb {
    ($which:expr, $callback:expr) => (install_cb($which, Some(Box::new($callback))))
}

// Called when the wasm module is instantiated
#[wasm_bindgen(start)]
pub fn main() -> Result<(), JsValue> {

    let state_ref = Rc::new(RefCell::new(get_document()?));
    let mut init_state = state_ref.borrow_mut();
    {
        // Animation callback aka main loop
        let state = state_ref.clone();
        init_state.install_animation_cb(Box::new(move || {
            let state : &mut State = &mut state.borrow_mut();
            state.frame_n += 1;
            let time = state.performance.now() as u64;
            state.renderer.render(&mut state.game, time);
            state.fps.set_text_content(Some(format!("frame: {}", state.frame_n).as_str()));
            state.renew_animation_cb();
        }));
    }
    {
        let state = state_ref.clone();
        install_cb!(&|x| init_state.window.set_onmousemove(x), move |event| {
            let state : &mut State = &mut state.borrow_mut();
			let event = event.dyn_into::<web_sys::MouseEvent>().unwrap();
            let x = event.page_x();
            let y = event.page_y();
        });
    }
    {
        let state = state_ref.clone();
        install_cb!(&|x| init_state.window.set_onmousedown(x), move |event| {
            let state : &mut State = &mut state.borrow_mut();
			let event = event.dyn_into::<web_sys::MouseEvent>().unwrap();
            let x = event.page_x();
            let y = event.page_y();
            state.game.on_mouse_click(&mut state.renderer, (x as f32, y as f32));
        });
    }
    {
        let state = state_ref.clone();
        install_cb!(&|x| init_state.window.set_onresize(x), move |event| {
            let state : &mut State = &mut state.borrow_mut();
            state.renderer.update_size(&state.canvas);
        });
    }
    {
        let state = state_ref.clone();
        install_cb!(&|x| init_state.window.set_onkeydown(x), move |event| {
            let state : &mut State = &mut state.borrow_mut();
			let event = event.dyn_into::<web_sys::KeyboardEvent>().unwrap();
            let key = event.key();
            match key.as_str() {
                "a" => state.game.user_input(Some(game::Direction::Left)),
                "d" => state.game.user_input(Some(game::Direction::Right)),
                "s" => state.game.user_input(Some(game::Direction::Down)),
                "w" => state.game.user_input(Some(game::Direction::Up)),
                _ => {},
            }
            state.last_input = key;
        });
    }
    {
        let state = state_ref.clone();
        install_cb!(&|x| init_state.window.set_onkeyup(x), move |event| {
            let state : &mut State = &mut state.borrow_mut();
			let event = event.dyn_into::<web_sys::KeyboardEvent>().unwrap();
            let key = event.key();
            if key == state.last_input {
                state.game.user_input(None);
            }
        });
    }

    Ok(())
}

#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(js_namespace = console)]
    fn log(s: &str);

    #[wasm_bindgen(js_namespace = console, js_name = log)]
    fn log_js(a: &wasm_bindgen::JsValue);

    #[wasm_bindgen(js_namespace = JSON, js_name = parse)]
    fn json_parse(a: &str) -> wasm_bindgen::JsValue;
}

struct State {
    frame_n : u32,
    window : web_sys::Window,
    document : web_sys::Document,
    body: web_sys::HtmlElement,
    fps: web_sys::HtmlDivElement,
    canvas: web_sys::HtmlCanvasElement,
    refresh_cb : Option<Closure<dyn Fn()>>,
    renderer: renderer::Renderer,
    performance: web_sys::Performance,
    game : game::Game,
    last_input : String,
}

impl State {
    // called once on initialization
    fn install_animation_cb(&mut self, cb: Box<dyn Fn()>) {
        self.refresh_cb = Some(Closure::wrap(cb));
        self.renew_animation_cb();
    }

    // called each frame to request the next callback
    fn renew_animation_cb(&self) {
        if let Some(cb) = self.refresh_cb.as_ref() {
            self.window.request_animation_frame(cb.as_ref().unchecked_ref()).unwrap();
        }
    }
}

// Installs a closure as a callback function.
// Leaks the callback closure to keep it alive, so use sparingly!
fn install_cb (
        setter : &dyn Fn(Option<&js_sys::Function>) -> (),
        callback : Option<Box<dyn FnMut(JsValue) -> ()>>) {
    if let Some(cb) = callback {
        let closure = Closure::wrap(cb);
        let fun = closure.as_ref().unchecked_ref::<js_sys::Function>();
        setter(Some(&fun));
        closure.forget();
    } else {
        setter(None);
    }
}

fn get_document() -> Result<State, JsValue>
{
    let window = web_sys::window().ok_or("`window` missing")?;
    let document = window.document().ok_or("`document` missing")?;
    let body = document.body().ok_or("`body` missing")?;
    let fps = document.get_element_by_id("fps_counter")
        .ok_or("fps_counter missing")?
        .dyn_into::<web_sys::HtmlDivElement>()?;
    let canvas = document.get_element_by_id("canvas")
        .ok_or("canvas missing")?
        .dyn_into::<web_sys::HtmlCanvasElement>()?;
    let renderer = renderer::Renderer::init(&canvas)?;
    let performance = window.performance().ok_or("No performance object")?;
    let png_image = include_bytes!("level.png");
    let game_data = decode_png(png_image).ok_or("Decode failed")?;
    let game = game::Game::new(&game_data, 128, 128).ok_or("Failed to init game")?;
    let last_input = String::new();
    Ok (State{
        frame_n : 0,
        window,
        document,
        body,
        fps,
        canvas,
        refresh_cb : None,
        renderer,
        performance,
        game,
        last_input,
    })
}

fn decode_png(data: &[u8]) -> Option<Vec::<u8>> {
    let mut decoder = png::Decoder::new(data);
    decoder.set_transformations(png::Transformations::IDENTITY);
    let (info, mut reader) = decoder.read_info().ok()?;
    let mut img_data = vec![0;info.buffer_size()];
    reader.next_frame(&mut img_data).ok()?;
    Some(img_data)
}
