use super::*;

pub struct Renderer {
    gl : Gl,
    width : u32,
    height : u32,
    vertex_shader : GlShader,
    fragment_shader : GlShader,
    shader_program : GlProgram,
    attrib_locations : RenderLocations,
    vertex_buffer : GlBuffer,
    texture : Option<GlTexture>,
    current_view : CameraView,
    vertices: std::vec::Vec<Vertex>,
}

#[derive(Copy, Clone)]
pub struct CameraView {
    view_radius : f32,
    view_center : glm::Vec2,
}

impl CameraView {
    pub fn new(radius : f32, center : (f32, f32)) -> Self {
        CameraView{
            view_radius : radius,
            view_center : glm::vec2(center.0, center.1),
        }
    }
    pub fn blend(a : CameraView, b : CameraView, t : f32) -> Self {
        let t_inv = 1.0 - t;
        CameraView {
            view_radius : a.view_radius*t + b.view_radius*t_inv,
            view_center : a.view_center*t + b.view_center*t_inv,
        }
    }
    pub fn apply(&self, w : u32, h : u32) -> glm::Mat4 {
        let w = w as f32;
        let h = h as f32;
        let smaller = if w < h {w} else {h};
        let scale_coeff = self.view_radius / smaller;
        let tx = (self.view_center.x*16.0).round()/16.0;
        let ty = (self.view_center.y*16.0).round()/16.0;
        let view_radius = (self.view_radius*16.0).round()/16.0;
        let top = view_radius * w / smaller;
        let right = view_radius * h / smaller;
        let move_model = glm::translation(&glm::vec3(-tx, -ty, 0.0));
        let projection = glm::ortho(-top, top, -right, right, -1.0, 1.0);
        projection * move_model
    }
}

impl Renderer {
    pub fn init(canvas : &web_sys::HtmlCanvasElement) -> Result<Self, JsValue> {
        // holy crap, why is the API so retarded I need to do stuff like that?
        let options = json_parse(r#"
            {
                "antialias" : false,
                "depth" : false,
                "stencil" : false
            }
            "#);
        log_js(&options);
        let gl = canvas.get_context_with_context_options("webgl", &options)?
            .ok_or("No webgl gl")?
            .dyn_into::<Gl>()?;
        let vertex_shader = shader::compile_shader(&gl, Gl::VERTEX_SHADER, r#"
            precision highp float;
            uniform mat4 view_transform;
            attribute vec3 position;
            attribute vec2 tex_coords;
            attribute vec4 color;
            varying vec4 tint_color;
            varying vec2 fragment_tex_coords;
            void main()
            {
                gl_Position = view_transform * vec4(position,1.0);
                fragment_tex_coords = tex_coords * (16.0/512.0);
                //tint_color = color;
                tint_color = pow(color, vec4(1.0/2.2));
            }
            "#)?;
        let fragment_shader = shader::compile_shader(&gl, Gl::FRAGMENT_SHADER, r#"
            precision mediump float;
            varying vec2 fragment_tex_coords;
            varying vec4 tint_color;
            uniform sampler2D texture;
            void main() {
                vec4 tex_color = texture2D(texture, fragment_tex_coords);
                if (tex_color.a == 0.0) {
                    discard;
                }
                //vec4 tex_color_linear = pow(tex_color, vec4(1.0/2.2));
                //vec4 tint_color_linear = pow(tint_color, vec4(1.0/2.2));
                //gl_FragColor = pow(tex_color_linear*tint_color_linear, vec4(2.2));
                gl_FragColor = tex_color * tint_color;
                //gl_FragColor = tex_color * pow(tint_color, vec4(1.0/1.5));
            }
            "#)?;
        let shader_program = shader::link_program(&gl, &vertex_shader, &fragment_shader)?;
        let attrib_locations = RenderLocations::query_locations(&gl, &shader_program);
        let vertex_buffer = gl.create_buffer().ok_or("Could not create buffer")?;
        let texture = gl.create_texture();
        gl.active_texture(Gl::TEXTURE0);
        gl.bind_texture(Gl::TEXTURE_2D, texture.as_ref());
        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MIN_FILTER, Gl::NEAREST as i32);
        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MAG_FILTER, Gl::NEAREST as i32);
        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_S, Gl::REPEAT as i32);
        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_T, Gl::REPEAT as i32);
        let png_image = include_bytes!("sprites.png");
        let pixel_data = decode_png(png_image).ok_or("Decode failed")?;
        gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
            Gl::TEXTURE_2D /*target*/,
            0 /*level*/,
            Gl::RGBA as i32 /*internalformat*/,
            512 as i32 /*width*/,
            512 as i32 /*height*/,
            0 /*border*/,
            Gl::RGBA /*format*/,
            Gl::UNSIGNED_BYTE /*type*/,
            Some(pixel_data.as_slice()) /*pixels*/,
        )?;
        let mut renderer = Renderer {
            gl,
            width: 0,
            height: 0,
            vertex_shader,
            fragment_shader,
            shader_program,
            attrib_locations,
            vertex_buffer,
            texture,
            current_view : CameraView::new(32.0, (8.0, 8.0)),
            vertices: Vec::new()};
        renderer.update_size(canvas);
        Ok(renderer)
    }

    pub fn update_size(&mut self, canvas : &web_sys::HtmlCanvasElement) {
        let width = canvas.scroll_width() as u32;
        let height = canvas.scroll_height() as u32;
        canvas.set_width(width);
        canvas.set_height(height);
        self.width = width;
        self.height = height;
        self.gl.viewport(0, 0, width as i32, height as i32);
    }

    pub fn render(&mut self, game : &mut game::Game, time: u64) {
        self.vertices.clear();
        let (new_view, sluggishness) = game.render(&mut self.vertices, time);
        self.current_view = CameraView::blend(self.current_view, new_view, sluggishness);
        let gl = &self.gl;
        gl.clear_color(0.0, 0.0, 0.0, 1.0);
        gl.clear(Gl::COLOR_BUFFER_BIT);
        gl.use_program(Some(&self.shader_program));
        let view_transform : glm::Mat4 = self.current_view.apply(self.width, self.height);
        gl.uniform_matrix4fv_with_f32_array(
            self.attrib_locations.view_transform_unif.as_ref(),
            false,
            glm::value_ptr(&view_transform));
        gl.uniform1i(self.attrib_locations.texture_unif.as_ref(), 0);
        upload_vertex_data(gl, &self.attrib_locations, &self.vertex_buffer, &self.vertices);
        gl.draw_arrays(Gl::TRIANGLES, 0, self.vertices.len() as i32);
        self.vertices.clear();
        game.render_ui(&mut self.vertices);
        let ui_transform = glm::ortho(0.0, self.width as f32 / 4.0, 0.0, self.height as f32 / 4.0, -1.0, 1.0);
        gl.uniform_matrix4fv_with_f32_array(
            self.attrib_locations.view_transform_unif.as_ref(),
            false,
            glm::value_ptr(&ui_transform));
        upload_vertex_data(gl, &self.attrib_locations, &self.vertex_buffer, &self.vertices);
        gl.draw_arrays(Gl::TRIANGLES, 0, self.vertices.len() as i32);
        gl.flush();
    }

    pub fn model_position_from_screen_coords(&self, position : (f32, f32)) -> (f32, f32) {
        let transform = self.current_view.apply(self.width, self.height);
        let inverse = glm::affine_inverse(transform);
        let (x, y) = position;
        // flip y-axis
        let y = self.height as f32 - y;
        // map (0, width) to (-1, 1)
        let x = 2.0*x / self.width as f32 - 1.0;
        let y = 2.0*y / self.height as f32 - 1.0;
        let v = glm::vec4(x, y, 0.0, 1.0);
        let result = inverse * v;
        (result.x, result.y)
    }
}

#[repr(C)]
#[derive(Copy,Clone)]
pub struct Vertex {
    pos : glm::U16Vec3,
    tex_coords : glm::U8Vec2,
    color : glm::U8Vec4,
}

impl Vertex {
    pub fn new() -> Self {
        Vertex {
            pos : glm::vec3(0, 0, 0),
            color : glm::vec4(0, 0, 0, 0),
            tex_coords : glm::vec2(0, 0)}
    }
}

struct RenderLocations {
    view_transform_unif : Option<GlUniformLocation>,
    texture_unif : Option<GlUniformLocation>,
    position_attrib : Option<u32>,
    tex_coords_attrib : Option<u32>,
    color_attrib : Option<u32>,
}

fn get_attrib_location(gl: &Gl, prog: &GlProgram, name: &str) -> Option<u32> {
    let location = gl.get_attrib_location(prog, name);
    if location == -1 {
        None
    } else {
        Some(location as u32)
    }
}

impl RenderLocations {
    fn query_locations(gl: &Gl, prog: &GlProgram) -> Self {
        let view_transform_unif = gl.get_uniform_location(prog, "view_transform");
        let texture_unif = gl.get_uniform_location(prog, "texture");
        let position_attrib = get_attrib_location(gl, prog, "position");
        let tex_coords_attrib = get_attrib_location(gl, prog, "tex_coords");
        let color_attrib = get_attrib_location(gl, prog, "color");
        RenderLocations {
            view_transform_unif,
            texture_unif,
            position_attrib,
            tex_coords_attrib,
            color_attrib,
        }
    }
}

pub fn define_rect(
        data : &mut std::vec::Vec<Vertex>,
        position : (u16, u16),
        tile : (u8, u8),
        brightness : (u8, u8, u8, u8)) {
    let (lx, ly) = position;
    let (tx, ty) = tile;
    let hx = lx + 16;
    let hy = ly + 16;
    let tlx = tx;
    let thx = tx + 1;
    let tly = ty + 1;
    let thy = ty;
    let (bl, tl, tr, br) = brightness;
    // lower left corner
    let llc = Vertex {
        pos : glm::vec3(lx, ly, 0),
        color : glm::vec4(bl, bl, bl, 255),
        tex_coords : glm::vec2(tlx, tly)};
    // lower right corner
    let lrc = Vertex {
        pos : glm::vec3(hx, ly, 0),
        color : glm::vec4(br, br, br, 255),
        tex_coords : glm::vec2(thx, tly)};
    // upper left corner
    let ulc = Vertex {
        pos : glm::vec3(lx, hy, 0),
        color : glm::vec4(tl, tl, tl, 255),
        tex_coords : glm::vec2(tlx, thy)};
    // upper right corner
    let urc = Vertex {
        pos : glm::vec3(hx, hy, 0),
        color : glm::vec4(tr, tr, tr, 255),
        tex_coords : glm::vec2(thx, thy)};

    let difference = |a, b| {if a > b {a - b} else {b - a}};
    let diag1 = difference(bl, tr);
    let diag2 = difference(tl, br);
    if diag1 > diag2 {
        data.push(llc);
        data.push(ulc);
        data.push(lrc);
        data.push(urc);
        data.push(ulc);
        data.push(lrc);
    } else {
        data.push(ulc);
        data.push(llc);
        data.push(urc);
        data.push(lrc);
        data.push(urc);
        data.push(llc);
    }
}

fn upload_vertex_data(
        gl: &Gl,
        locations: &RenderLocations,
        buffer: &GlBuffer,
        data: &[Vertex]) {

    gl.bind_buffer(Gl::ARRAY_BUFFER, Some(buffer));
    let vertex_size = std::mem::size_of::<Vertex>();
    let data_len = data.len();
    let total_len = data_len * vertex_size;
    let data8_ptr : *const u8 = data.as_ptr() as *const u8;
    unsafe {
        let data8 : &[u8] = std::slice::from_raw_parts(data8_ptr, total_len);
        gl.buffer_data_with_u8_array(Gl::ARRAY_BUFFER, data8, Gl::DYNAMIC_DRAW);
    }
    if let Some(position_attrib) = locations.position_attrib {
        gl.vertex_attrib_pointer_with_i32(
            position_attrib /*index*/,
            3 /*size*/,
            Gl::UNSIGNED_SHORT /*type*/,
            false /*normalzied*/,
            vertex_size as i32 /*stride*/,
            offset_of!(Vertex => pos).get_byte_offset() as i32 /*offset*/
        );
        gl.enable_vertex_attrib_array(position_attrib);
    }
    if let Some(color_attrib) = locations.color_attrib {
        gl.vertex_attrib_pointer_with_i32(
            color_attrib /*index*/,
            4 /*size*/,
            Gl::UNSIGNED_BYTE /*type*/,
            true /*normalzied*/,
            vertex_size as i32 /*stride*/,
            offset_of!(Vertex => color).get_byte_offset() as i32 /*offset*/
        );
        gl.enable_vertex_attrib_array(color_attrib);
    }
    if let Some(tex_coords_attrib) = locations.tex_coords_attrib {
        gl.vertex_attrib_pointer_with_i32(
            tex_coords_attrib /*index*/,
            2 /*size*/,
            Gl::UNSIGNED_BYTE /*type*/,
            false /*normalzied*/,
            vertex_size as i32 /*stride*/,
            offset_of!(Vertex => tex_coords).get_byte_offset() as i32 /*offset*/
        );
        gl.enable_vertex_attrib_array(tex_coords_attrib);
    }
}

